//--------------------------------------------------------------------------------------------//
// This is the main Application File for the ITC2007 competition                              //                                 
// Author : R Lewis                                                                           //
// Date last modified : 17/09/2013                                                           //
//--------------------------------------------------------------------------------------------//

#include "application.h"

//-------------------------------------------------------------------------------------
//	Arrays and variables used throughout the program
//-------------------------------------------------------------------------------------
int *roomSize, *eventSize, *totalNumConflict, **before, *originalTotalNumPlaces;
bool **attends, **roomFeatures, **eventRoom, **conflict, **eventFeatures, **originalEventPlace, **eventAvail, **roomAvail, *hasPrecConstr;
int numEvents, numRooms, numFeatures, numStudents, NUMBEROFPLACES, numDummyRooms, weightCoefficient, maxNumKempeChains;
bool outputOn;
int flexibility;
TwoDIntVector precidence, beforePrec, feasRooms, eventStus;
IntVector NumKempeChainsRecord;
unsigned long long propMoves, feasMoves;

void usage() {
	cout<<"Two Stage Algorithm for ITC 2007\n";
	cout<<"--------------------------------\n\n";
	cout<<"Usage:\n"
	<<"-t <int>                     (timelimit. Default 10s)\n"
	<<"-r <int>                     (random seed. Default 1)\n"
	<<"-f <int>                     (flexibility level: (1) noMaxmatch, (2) maxMatch, (3) maxMatchandKempe, (4) maxMatchDoubleKempeChain, (5) maxMatchTripleKempeChain,... etc)\n"
	<<"-e <double>                  (simulated annealing end temperature. Default = 0.5)\n"
	<<"---------------------------------\n"
	<<"-DR <int>                    (number of dummy rooms. Default = 0)\n"
	<<"-W <int>                     (weight coefficient (if dummy rooms > 0). Default = 2)\n"
	<<"---------------------------------\n"
	<<"-R                           (if present, do a random walk (no optimisation) to get a good estimate of the Feasibility Ratio)\n"
	<<"-RP <int>                    (partialCol is refreshed whenever <int> iterations have occurred without an improvement. Default = 5000; Negative numbers mean infinity)\n"
	<<"-o                           (if present, output is set to on)\n"
	<<"-l                           (if present, legacy .tim files (from ITC 2003 or 'hard' instances) are assumed, else ITC2007 instances are assumed)\n"
	<<"---------------------------------\n"
	<<"inputFile                    (Name of the input file (no extension). This is the only compulsory argument and should come last\n\n";
	
	exit(0);
}


int main(int argc, char**argv){

	if(argc<=1){
		usage();
	}
	
	//Variables (default values)
	char fileName[40], slnFileName[40];
	clock_t timeLimit = 10;
	int seed = 1, i, pColRefreshRate = 5000;
	bool doRandomWalk = false, legacyTim = false, pColRefresh = true;
	double SAEndTemp=0.5;
	outputOn = false;
	flexibility = 3;
	numDummyRooms = 0;
	propMoves = 0;
	weightCoefficient = 2;
	feasMoves = 0;
	ifstream inStream; 
	
	//Read in program arguments
	for (i=1; i<argc; i++) {
		if (strcmp("-t", argv[i]) == 0)			timeLimit = atoi(argv[++i]);
		else if (strcmp("-r", argv[i]) == 0)	seed = atoi(argv[++i]);
		else if (strcmp("-f", argv[i]) == 0)	flexibility = atoi(argv[++i]);
		else if (strcmp("-e", argv[i]) == 0)	SAEndTemp = atof(argv[++i]);
		else if (strcmp("-R", argv[i]) == 0)	doRandomWalk = true;
		else if (strcmp("-RP", argv[i]) == 0)	pColRefreshRate = atoi(argv[++i]);
		else if (strcmp("-W", argv[i]) == 0)	weightCoefficient = atoi(argv[++i]);
		else if (strcmp("-DR", argv[i]) == 0)	numDummyRooms = atoi(argv[++i]);
		else if (strcmp("-o", argv[i]) == 0)	outputOn = true;
		else {
			if (strcmp("-l", argv[i]) == 0 )	{legacyTim = true; i++;}
			strcpy(fileName,argv[i]);
			strcpy(slnFileName,argv[i]);
			strcat(fileName,".tim");
			inStream.open(fileName);
			readTimFileAndPreProcess(inStream,legacyTim);
			inStream.close();
		}
	}
	
	//Set up time limit, random seed, and output
	srand(seed);
	maxNumKempeChains = flexibility - 2;
	if(pColRefreshRate <= 0) {pColRefresh = false; pColRefreshRate = -1;}
	NumKempeChainsRecord.resize(maxNumKempeChains+1,0);
	if(outputOn)cout<<"Doing experiment with '"<<fileName<<"'. Seed "<<seed<<endl<<"numEvents = "<<numEvents<<", numRooms = "<<numRooms<<"("<<numDummyRooms<<"), numStudents = "<<numStudents<<", flexibility = "<<flexibility<<", W = "<<weightCoefficient<<endl;

	//Now start the clock and set the various timelimits (stage 1 run for a maximum of 50% of the time limit)
	clock_t clockStart = clock(); 
	time_t stime, ftime, etime;
	time(&stime);
	clock_t Part1ClockFinish = clockStart + int(0.5*timeLimit*CLOCKS_PER_SEC);
	clock_t Part2ClockFinish = Part1ClockFinish + int(0.5*timeLimit*CLOCKS_PER_SEC);
		
	if(outputOn){
		cout<<"TimeLimit = "<<timeLimit<<endl;
		cout<<"Part 1 will be run for (max) "<<(Part1ClockFinish-clockStart)/CLOCKS_PER_SEC<<" seconds\n";
		cout<<"Part 2 will be run for (min) "<<(Part2ClockFinish-Part1ClockFinish)/CLOCKS_PER_SEC<<" seconds\n";
	}

	//---------------------------------------------------------------------------------------
	//                             INITIAL SOL GENERATION
	//---------------------------------------------------------------------------------------
	TTAndUnplaced theSolution = generateNewSolution();
	int evsUnplacedAtStart = theSolution.Unplaced.size(), pColIts=0;

	//---------------------------------------------------------------------------------------
	//                             PARTIAL COL
	//---------------------------------------------------------------------------------------
	clock_t theClockMarker;
	theClockMarker = clock();

	if(theSolution.Unplaced.size() > 0){
		pColIts = doPartialCol(theSolution,Part1ClockFinish,pColRefresh,pColRefreshRate);
		if(theSolution.Unplaced.size() > 0 && outputOn) cout<<"Time limit for PartialCol reached. There are still "<<theSolution.Unplaced.size()<<" unplaced events.\n";
	}
	int pColUnplaced = theSolution.Unplaced.size();
	double pColTime = (double)(clock()-clockStart)/CLOCKS_PER_SEC;

	//---------------------------------------------------------------------------------------
	//                    SA BIT  FOR THE SOFT CONSTRAINTS                                  
	//---------------------------------------------------------------------------------------
	//if we are here then we have completed stage 1 and will have a timetable that obeys all hard constraints, but which may still have some unplaced events.
	//We will now set about trying to satisfy the soft constraints of the timetable.
	
	//Evaluate the timetable according to the soft constraints by using the studentavailability matrix and corresponding days matrix
	int softCost = 0, startSoftCost, dayCosts[NUMDAYS], **stuDayCosts, dToF=calcDistanceToFeasibility(theSolution), bestDToF=dToF, chainCnt=0;
	bool perfectSolutionFound = false, **studentavailability;
	
	//Set up the arrays
	stuDayCosts = new int* [NUMDAYS];
	for(i=0; i<NUMDAYS; i++) stuDayCosts[i] = new int [numStudents];
	studentavailability = new bool* [NUMBEROFSLOTS];
	for (i=0;i<NUMBEROFSLOTS;i++) studentavailability[i] = new bool[numStudents];
	
	//And fill them according to the solution we have
	makeSAvailFromSol(theSolution,studentavailability);
	makeStuDayCosts(studentavailability,stuDayCosts);
	fillDayCosts(stuDayCosts, dayCosts);
	softCost = totalStuCosts(stuDayCosts);
	startSoftCost = softCost;
	
	//Check to see the state of theSolution
	if(outputOn)cout<<"\nThe initial feasible timetable has "<<softCost<<" soft constraint violations (DToF = "<<dToF<<")"<<endl;
	if(dToF == 0 && softCost == 0) perfectSolutionFound = true;

	//We now anneal theSolution making sure that all of the hard constraints are never violated in the timetable
	if(doRandomWalk){
		doFeasRandWalk(softCost,Part2ClockFinish,theSolution,studentavailability,stuDayCosts,dayCosts);
	}
	else if(!perfectSolutionFound){
		doSAPhase2(softCost,dToF,Part2ClockFinish,theSolution,studentavailability,stuDayCosts,dayCosts,SAEndTemp,chainCnt);
	}

	//Remove dummyRooms and contents if necesarry
	if(numDummyRooms > 0) eraseDummyRooms(theSolution,softCost);
		
	//Error checking
	checkFeasibility(theSolution);
	
	//---------------------------------------------------------------------------------------
	//OUTPUT EXPERIMENT INFORMATION
	outputTimAnswerFile(theSolution,slnFileName,seed);
	if(outputOn)cout<<"There are "<<theSolution.Unplaced.size()<<" events that are unplaced in the final solution "<<endl;
	if(outputOn)cout<<"The resultant Distance to Feasibility is "<<calcDistanceToFeasibility(theSolution)<<endl;
	if(outputOn)cout<<"The soft cost of this timetable is "<<softCost<<endl;
	int finalDtF = calcDistanceToFeasibility(theSolution);
		  
	//---------------------------------------------------------------------------------------
	time(&ftime);
	etime = ftime-stime;
	if(outputOn)cout<<"Wall clock time = "<<etime<<endl;
	double duration = (double)(clock()-clockStart)/CLOCKS_PER_SEC;

	//---------------------------------------------------------------------------------------
	//WRITE SOME DETAILS ABOUT THE RUN TO THE MASTER LOG FILE
	ofstream resultsLog("resultsLog.txt", ios::app);
	resultsLog	<<fileName<<"\t"<<numEvents<<"\t"<<numRooms<<"\t"<<numStudents<<"\t"<<timeLimit<<"\t"<<seed<<"\t"<<flexibility<<"\t"<<numDummyRooms<<"\t";	//Run Conditions
	if(numDummyRooms <= 0) resultsLog<<"n/a"<<"\t";
	else				   resultsLog<<weightCoefficient<<"\t";
	resultsLog	<<evsUnplacedAtStart<<"\t"<<pColIts<<"\t"<<pColTime<<"\t"											//PartialCol Performance
				<<theSolution.Unplaced.size()<<"\t"<<finalDtF<<"\t"													//Final Solution DtF
				<<doRandomWalk<<"\t"<<chainCnt<<"\t"<<feasMoves/double(propMoves)<<"\t"<<SAEndTemp<<"\t"<<startSoftCost<<"\t"<<softCost<<"\t"<<duration<<"\n";		//Feas Ratio and SA Performance
	resultsLog.close();
	
	//---------------------------------------------------------------------------------------
	freeUpHeap();
	if(outputOn)cout << "Program ended. "<<duration<<" seconds to of CPU time to complete\n";
	return 0;
}

//----------------------------------------------------------------------------------------
inline
void outputTimAnswerFile(TTAndUnplaced &theSolution, char *slnfilename, int seed)
{	
	//Outputs the best solution acheived in the desired format for the checker program
	char outfileName[50];
	strcpy(outfileName, slnfilename);
	strcat(outfileName, ".sln\0");	
	ofstream ansStream;
	ansStream.open(outfileName);
	IntVector event_rooms(numEvents,-1), event_slots(numEvents,-1); 
	int r, c, i;
	for(r=0; r<numRooms; r++){
		for(c=0; c<NUMBEROFSLOTS; c++){
			if(theSolution.TT[r][c]!= -1){
				event_rooms[theSolution.TT[r][c]] = r;
				event_slots[theSolution.TT[r][c]] = c;
			}
		}
	}
	for(i=0; i<numEvents; i++)ansStream<<event_slots[i]<<' '<<event_rooms[i]<<endl;
	ansStream.close();
}

//----------------------------------------------------------------------------------------
inline
int calcDistanceToFeasibility(TTAndUnplaced &theSolution)
{
	int i, distance = 0;
	for(i=0; i<theSolution.Unplaced.size(); i++)distance += eventSize[theSolution.Unplaced[i]];
	return distance;
}


