#include "partialcol.h"

extern int *roomSize, *eventSize, *totalNumConflict, **before, *originalTotalNumPlaces;
extern bool **attends, **roomFeatures, **eventRoom, **conflict, **eventFeatures, **originalEventPlace, **eventAvail, **roomAvail, *hasPrecConstr;
extern int numEvents, numRooms, numFeatures, numStudents, NUMBEROFPLACES;
extern bool outputOn;
extern TwoDIntVector precidence, beforePrec, feasRooms, eventStus;


//**************************************************************************************
inline 
void maxMatch(int evToInsert, int j, TTAndUnplaced &sol, bool &done, IntVector &col, IntVector &pred, IntQueue &Q)
{
	//Maximum matching algorithm based on Bredth first search. Given a "feasible" timeslot j, we have an event to insert
	//that is known not to clash with any events in the timeslot, but is not suitable for any of the available rooms.
	//Thus the procedure below is run. If it is succesful, then the changes are made and done = true; 
	//otherwise, no changes are made and done is false.
	
	int endPoint,r,ev,rm;	
	
	//initialise data structures needed for BFS
	col.clear();
	col.resize(numRooms*2+1, WHITE);
	pred.clear();
	pred.resize(numRooms*2+1, NIL);
	Q.clear();
	col.back() = GREY;
	Q.push_back(numRooms*2);
	
	while(!Q.empty()){
		int u = Q.front();
		if(u>=numRooms){
			//We are currently at an "event node" u. Check to see if there is a room available for it
			//First determine the event "ev" we are dealing with
			if(u==numRooms*2) 
				ev = evToInsert;
			else 
				ev = sol.TT[u-numRooms][j];
			for(r=0;r<feasRooms[ev].size();r++){
				rm = feasRooms[ev][r];
				if(sol.TT[rm][j]==-1){
					//Room "rm" is suitable and available for event u so we end
					pred[rm] = u;
					endPoint = rm;
					goto DONE;
				}
				if(col[rm]==WHITE){
					//Room "rm" is suitable, but not available, so we will explore from it
					col[rm]=GREY;
					pred[rm] = u;
					Q.push_back(rm);
				}
			}
		}
		else {
			//We are currently at a room node u that is occupied, so we traverse to the event node that is occupying it
			int v = u+numRooms;
			if(col[v]==WHITE){
				col[v]=GREY;
				pred[v] = u;
				Q.push_back(v);
			}
		}
		Q.pop_front();
		col[u] = BLACK;
	}
	//If we are here then a matching is not possible.
	return;

DONE:
	//If we are here then a matching was found. Thus we now make the changes to the timetable
	int current = endPoint;
	while(pred[pred[current]] != NIL){
		sol.TT[current][j] = sol.TT[pred[current]-numRooms][j];
		current = pred[pred[current]];
	}
	sol.TT[current][j] = evToInsert;
	done=true;
}

//----------------------------------------------------------------------------------------------
inline 
int minVal(int a, int b){
	if(a<b) return a;
	else return b;
}
inline
bool moveTabu(int ev, int slot, int it, TwoDIntVector &tabuList){
	//Returns true if inserting "ev" into "slot" is tabu, else false.
	if(tabuList[ev][slot] > it) return true;
	else return false;
}
inline 
void makeTabu(TwoDIntVector &tabuList, int ev, int slot, int it, int cost){
	tabuList[ev][slot] = it + (int(0.6*cost))+(rand()%10);	
}
inline
bool eventFeasForSlot(int ev, int slot, IntVector &slotOfEvent)
{
	//Checks to see if inserting ev into slot violates any eventAvail or prec constraints. Return false if it does.
	if(!eventAvail[ev][slot]) return false; // slot not available for ev
	if(!hasPrecConstr[ev]) return true; // ev is not associated with any prec constraint, so we can end.
	
	//If we are here, ev is involved with a precidence constraint, so we need to check the slots of the associated events
	int i;
	for(i=0; i<precidence[ev].size(); i++){
		//Check timeslots of all events that have to come after ev 
		if(slotOfEvent[precidence[ev][i]] != -1) {
			if(slotOfEvent[precidence[ev][i]] <= slot) return false; 
		}
	}
	for(i=0; i<beforePrec[ev].size(); i++){
		//Check timeslot of all events that have to come before ev 
		if(slotOfEvent[beforePrec[ev][i]] != -1) {
			if(slotOfEvent[beforePrec[ev][i]] >= slot) return false; 
		}
	}
	return true;
}
inline
void calcInsertionCost(TTAndUnplaced &sol, int ev, int slot, TwoDIntVector &eventSlotCost, TwoDBoolVector &eventSlotRmAvail, IntVector &tempSlot, 
						IntVector &col, IntVector &pred, IntQueue &Q, IntVector &slotOfEvent)
{
	if(!eventFeasForSlot(ev,slot,slotOfEvent)){
		//"Ev" cannot feasibly go into "slot" as it violates precidence and/or eventAvail constraints. 
		eventSlotCost[ev][slot] = INT_MAX;
		eventSlotRmAvail[ev][slot] = false;
		return;
	}
	//Otherwise, we need to ID the cost of inserting this event
	int r, clashes=0;
	bool foundRoom = false;
	//Calculate number of events that would need to be removed from "slot" to allow "ev" to be inserted
	for(r=0; r<numRooms; r++){
		if(sol.TT[r][slot] != -1){
			if(conflict[sol.TT[r][slot]][ev]){
				clashes++;
				if(eventRoom[ev][r]) foundRoom=true; //Removing this clashing events would free up a suitable room
			}
		} 
		else {
			if(eventRoom[ev][r]) foundRoom=true; //There is already a free room in this slot that is suitable
		}
	}
	if(!foundRoom){
		//If we are here, we need to remove the clashing events and see if we can then insert "ev" using the maximum matching algorithm. We first copy the slot and remove the clashing events
		for(r=0; r<numRooms; r++){
			tempSlot[r] = sol.TT[r][slot];
			if(sol.TT[r][slot] != -1 && conflict[sol.TT[r][slot]][ev]){
				sol.TT[r][slot] = -1;
			}
		}
		maxMatch(ev,slot,sol,foundRoom,col,pred,Q);
		//Whatever the result we now revert to the original
		for(r=0; r<numRooms; r++) sol.TT[r][slot] = tempSlot[r];
	}
	if(!foundRoom){
		//It turns out we will need to remove all of the clashing events plus one more to allow "ev" to be inserted into "slot"
		eventSlotCost[ev][slot] = clashes+1;
		eventSlotRmAvail[ev][slot] = false;
	}
	else {
		//We will only need to remove the clashing events to allow "ev" to be inserted into "slot"
		eventSlotCost[ev][slot] = clashes;
		eventSlotRmAvail[ev][slot] = true;
	}
}
inline
void doMove(TTAndUnplaced &sol, int ev, int slot, int pos, TwoDIntVector &eventSlotCost, TwoDBoolVector &eventSlotRmAvail, IntVector &tempSlot, IntVector &col, 
				IntVector &pred, IntQueue &Q, IntVector &slotOfEvent, TwoDIntVector &tabuList, int it, int bestQuality)
{
	//Inserts "ev" into "slot" and removes the appropriate events.
	int r, rm, i, j, numExtracted=0;
	bool inserted = false;
	//Remove clashing events from "slot", and add these to the unplaced list. 
	for(r=0;r<numRooms;r++){
		if(sol.TT[r][slot] != -1 && conflict[sol.TT[r][slot]][ev]){
			sol.Unplaced.push_back(sol.TT[r][slot]);
			numExtracted++;
			slotOfEvent[sol.TT[r][slot]] = -1;
			sol.TT[r][slot] = -1;
		}
	} 
	//Now add "ev" to the slot
	if(!eventSlotRmAvail[ev][slot]){
		//We have to remove one further event to free up a suitable room r. Ev can now be put here.
		r = feasRooms[ev][rand()%feasRooms[ev].size()];
		sol.Unplaced.push_back(sol.TT[r][slot]);
		numExtracted++;
		slotOfEvent[sol.TT[r][slot]] = -1;
		sol.TT[r][slot] = ev;
	}
	else {
		//We know that ev can definately be added to slot. So do it.
		for(r=0;r<feasRooms[ev].size();r++){
			rm = feasRooms[ev][r];
			if(sol.TT[rm][slot]==-1){
				sol.TT[rm][slot] = ev;
				break;
			}
		}
		if(r>=feasRooms[ev].size()){
			//If we are here, we need to run the maxMatch algorithm. We know that this should be succesful
			maxMatch(ev,slot,sol,inserted,col,pred,Q);
		}
	}
	//Make a note of "ev"s slot and remove "ev" from its position in the unplaced list
	slotOfEvent[ev] = slot;
	sol.Unplaced.erase(sol.Unplaced.begin() + pos);
	
	//Finally, we need to update the cost and tabu matrices. First look at the events in U that have not been effected
	for(i=0;i<sol.Unplaced.size()-numExtracted;i++){
		if(hasPrecConstr[sol.Unplaced[i]]){
			for(j=0;j<NUMBEROFSLOTS;j++){
				calcInsertionCost(sol,sol.Unplaced[i],j,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);
			}
		}
		else calcInsertionCost(sol,sol.Unplaced[i],slot,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);
	}
	//And then look at the events recently added to U
	for(i=sol.Unplaced.size()-numExtracted;i<sol.Unplaced.size();i++){
		for(j=0;j<NUMBEROFSLOTS;j++){
			calcInsertionCost(sol,sol.Unplaced[i],j,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);
		}
		//All newly extracted events are forbidden to go back into "slot" for a specific number of iterations
		makeTabu(tabuList,sol.Unplaced[i],slot,it,sol.Unplaced.size());
	}
}
inline
void doRandomMove(TTAndUnplaced &sol, TwoDIntVector &eventSlotCost, TwoDBoolVector &eventSlotRmAvail, IntVector &slotOfEvent, TwoDIntVector &tabuList, int it,
				IntVector &tempSlot, IntVector &col, IntVector &pred, IntQueue &Q, int bestQuality){
	//Randonly select an event in the timetable, un-place it
	int rm,slot,i,j;
	do {
		rm = rand()%numRooms;
		slot = rand()%NUMBEROFSLOTS;
	} while(sol.TT[rm][slot]==-1);
	sol.Unplaced.push_back(sol.TT[rm][slot]);
	slotOfEvent[sol.TT[rm][slot]] = -1;
	sol.TT[rm][slot] = -1;
	//Mark this move as tabu and update the cost matrices
	for(i=0;i<sol.Unplaced.size()-1;i++){
		if(hasPrecConstr[sol.Unplaced[i]]){
			for(j=0;j<NUMBEROFSLOTS;j++){
				calcInsertionCost(sol,sol.Unplaced[i],j,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);
			}
		}
		else calcInsertionCost(sol,sol.Unplaced[i],slot,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);
	}
	//And then look at the event just added to U
	for(j=0;j<NUMBEROFSLOTS;j++){
		calcInsertionCost(sol,sol.Unplaced.back(),j,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);
	}
	//The newly extracted event is forbidden to go back into "slot" for a specific number of iterations
	makeTabu(tabuList,sol.Unplaced.back(),slot,it,sol.Unplaced.size());
}
inline
void populateCostArray(TTAndUnplaced &sol, TwoDIntVector &eventSlotCost, TwoDBoolVector &eventSlotRmAvail, IntVector &tempSlot, IntVector &col, IntVector &pred, IntQueue &Q, IntVector &slotOfEvent)
{
	int i, j;
	for(i=0;i<sol.Unplaced.size();i++){
		for(j=0;j<NUMBEROFSLOTS;j++){
			calcInsertionCost(sol,sol.Unplaced[i],j,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);
		}
	}
}
inline
void refreshSearch(TTAndUnplaced &sol, IntVector &slotOfEvent, TwoDIntVector &eventSlotCost, TwoDBoolVector &eventSlotRmAvail, IntVector &tempSlot, IntVector &col, IntVector &pred, IntQueue &Q, TwoDIntVector &tabuList, int &it, clock_t Part1ClockFinish)
{
	int i, j;
	//First, extract 10% of events from the timetable
	if(outputOn) cout<<"Doing Refresh\n";
	for(i=0;i<numRooms;i++){
		for(j=0;j<NUMBEROFSLOTS;j++){
			if(sol.TT[i][j] != -1 && rand()%10 == 0){
				sol.Unplaced.push_back(sol.TT[i][j]);
				sol.TT[i][j] = -1;
			}
		}
	}
	//Now shuffle the other events remaining in the timetable
	doFeasRandWalkPCol(sol,slotOfEvent,Part1ClockFinish);
	populateCostArray(sol,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);
	//Finally, reset the tabu list
	it = 0;
	for(i=0;i<numEvents;i++)for(j=0;j<NUMBEROFSLOTS;j++)tabuList[i][j]=0;	
}

int doPartialCol(TTAndUnplaced &sol, clock_t Part1ClockFinish, bool pColRefresh, int pColRefreshRate)
{
	int ev, i, j, bestEv, bestSlot, besti, bestVal, numBest, it=0, totalIts=0, movesSinceImprovement=0;
	bool foundMove;

	if(outputOn)cout<<"\nRunning PartialCol (max "<<(Part1ClockFinish-clock())/CLOCKS_PER_SEC<<" seconds)...\n";

	//Declare and populate the vectors
	IntVector slotOfEvent(numEvents, -1), tempSlot(numRooms), col, pred;
	vector<vector<int> >  eventSlotCost(numEvents, vector<int>(NUMBEROFSLOTS));
	vector<vector<bool> > eventSlotRmAvail(numEvents, vector<bool>(NUMBEROFSLOTS));
	vector<vector<int> >  tabuList(numEvents, vector<int>(NUMBEROFSLOTS,0));
	IntQueue Q;
	//Put a copy of the current solution into a placeholder
	TTAndUnplaced bestSol = sol;
	
	//Populate slotOfEvent vector
	for(i=0; i<numRooms; i++){ 
		for(j=0; j<NUMBEROFSLOTS; j++){
			if(sol.TT[i][j] != -1) slotOfEvent[sol.TT[i][j]] = j;
		}
	}

	//Populate structure for holding details of each neighbourhood move
	populateCostArray(sol,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent);

	//Here is the tabu search algorithm
	while( (Part1ClockFinish > clock()) && (bestSol.Unplaced.size() > 0 ) ){
		//Now identify the best neighbourhood move
		bestVal = INT_MAX-1;
		numBest = 0;
		foundMove = false;
		for(i=0;i<sol.Unplaced.size();i++){
			ev = sol.Unplaced[i];
			for(j=0;j<NUMBEROFSLOTS;j++){
				if(eventSlotCost[ev][j] <= bestVal){
					if(eventSlotCost[ev][j]<bestVal) 
						numBest = 0;
					if(!moveTabu(ev,j,it,tabuList) || (sol.Unplaced.size()+eventSlotCost[ev][j]-1) < bestSol.Unplaced.size()){
						if(rand()%(numBest+1)==0){
							//Save this move
							bestVal = eventSlotCost[ev][j];
							bestEv = ev;
							bestSlot = j;
							besti = i;
							foundMove = true;
						}
						numBest++;
					}
				}
			}
		}
		if(bestVal >= INT_MAX-1){
			//No non-tabu (or feasible preserving) moves available. Randomly remove an event from the timetable
			doRandomMove(sol,eventSlotCost,eventSlotRmAvail,slotOfEvent,tabuList,it,tempSlot,col,pred,Q,bestSol.Unplaced.size());
		} 
		else {
			//Make the selected move (insert the best event into the best slot)
			doMove(sol,bestEv,bestSlot,besti,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,slotOfEvent,tabuList,it,bestSol.Unplaced.size());
			if(sol.Unplaced.size() < bestSol.Unplaced.size()){
				//Keep a note of the best solution and cost observed
				bestSol = sol;
				if(outputOn) cout<<"It: "<<setw(8)<<totalIts<<", |U| = "<<bestSol.Unplaced.size()<<"\n";
				movesSinceImprovement = 0;
			} else {
				movesSinceImprovement++;
			}
		}
		it++;
		totalIts++;
		if(movesSinceImprovement >= pColRefreshRate && pColRefresh){
			//No improvement has occurred for a long time. So randomly alter the solution to refresh the search (reset tabu list too)
			refreshSearch(sol,slotOfEvent,eventSlotCost,eventSlotRmAvail,tempSlot,col,pred,Q,tabuList,it,Part1ClockFinish);
			movesSinceImprovement = 0;
		}
	}
	//End of tabu search algorithm. Take the best observed solution and return number of iterations
	if(outputOn) cout<<"...Done\n";
	sol = bestSol;
	return totalIts;
}


